/*
 * A page mapper providing memory pages to satisfy file accesses.
 *
 * Copyright (C) 2021 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include "ipc.h"
#include "page_mapper.h"



PageMapper::PageMapper(Accessor *accessor, Pages *pages)
: _accessor(accessor), _pages(pages), _attached(0)
{
}

/* Accounting methods. */

/* Attach a pager, opening the accessor if required. */

void PageMapper::attach()
{
    std::lock_guard<std::mutex> guard(_lock);

    _attached += 1;
}

/* Detach a pager, purging active pages and closing the accessor if no more
   pagers are attached. Return whether any pagers are still attached. */

unsigned int PageMapper::detach()
{
    std::lock_guard<std::mutex> guard(_lock);

    if (_attached)
    {
        _attached -= 1;

        if (!_attached)
        {
            _map.purge(this, _pages);
            _accessor->close();
        }
    }

    return _attached;
}

/* Interface for the pager. */

/* Return a flexpage providing access to the indicated file 'offset'.

   The returned flexpage will either be an existing, compatible flexpage or a
   completely new flexpage.

   This method locks the mapper to prevent concurrent queries with the same
   details, with the lock held until the queue operation releases the lock. */

Flexpage *PageMapper::get(offset_t offset, flags_t flags)
{
    _lock.lock();

    Flexpage *f = find(offset);

    if (f == NULL)
        f = flexpage(offset);

    /* Record a new user of the flexpage and upgrade the access flags. */

    f->increment();
    f->upgrade(flags);
    return f;
}

/* Queue the given 'flexpage' in the page collection, making it available for
   eventual reuse.

   This method unlocks the mapper. */

void PageMapper::queue(Flexpage *flexpage)
{
    _pages->queue(this, flexpage);

    _lock.unlock();
}

/* Flush pages in the given range from 'start' with 'size'. */

void PageMapper::flush_all(offset_t start, offset_t size)
{
    std::lock_guard<std::mutex> guard(_lock);

    _map.flush_all(start, size, this, _pages);
}

/* Return the maximum extent of the mapped resource. */

offset_t PageMapper::get_data_size()
{
    return _accessor->get_size();
}

/* Set the maximum extent of the mapped resource. */

void PageMapper::set_data_size(offset_t size)
{
    _accessor->set_size(size);
}

/* Internal flexpage retrieval methods. */

/* Find an existing flexpage for 'offset'. Where the accessor has registered a
   compatible flexpage, an attempt is made to reserve it in the page collection;
   if this succeeds, the flexpage is returned. Otherwise, NULL is returned. */

Flexpage *PageMapper::find(offset_t offset)
{
    Flexpage *flexpage = _map.find(offset);

    /* Between finding and reserving a flexpage, there is a possibility that
       another accessor might acquire the flexpage, issue it, and even purge
       it. */

    if ((flexpage != NULL) && _pages->reserve(this, flexpage))
        return flexpage;
    else
        return NULL;
}

/* Obtain a new flexpage for the file 'offset'. If the page collection is unable
   to obtain a completely new flexpage, an existing flexpage is requested from
   the page collection and recycled.

   The obtained flexpage is filled with content. */

Flexpage *PageMapper::flexpage(offset_t offset)
{
    Flexpage *flexpage = _pages->flexpage();

    /* Obtain an existing flexpage and reuse it. */

    if (flexpage == NULL)
        flexpage = _pages->remove();

    flexpage->reset(offset);

    fill(flexpage);
    _map.insert(flexpage);
    return flexpage;
}

/* Interface for the page collection. */

/* Remove the record of 'flexpage' in this accessor, flushing its content. */

void PageMapper::remove(Flexpage *flexpage)
{
    _map.remove(this, flexpage);
}

/* Data transfer methods. */

void PageMapper::fill(Flexpage *flexpage)
{
    _accessor->fill(flexpage);
}

void PageMapper::flush(Flexpage *flexpage, bool purge)
{
    if (flexpage->decrement() || purge)
    {
        if (flexpage->modified())
            _accessor->flush(flexpage);

        ipc_unmap_flexpage(flexpage);
        flexpage->invalidate();
    }
}

// vim: tabstop=4 expandtab shiftwidth=4
